import os
import os.path
import re
import shutil

from pcs import settings
from pcs.common import reports
from pcs.common.reports.item import ReportItem
from pcs.common.str_tools import join_multilines
from pcs.lib.communication import qdevice_net as qdevice_net_com
from pcs.lib.communication.tools import run_and_raise
from pcs.lib.errors import LibraryError
from pcs.lib.tools import write_tmpfile

SERVICE_NAME = "corosync-qnetd"

__model = "net"
__qnetd_certutil = os.path.join(
    settings.corosync_qnet_binaries, "corosync-qnetd-certutil"
)
__qnetd_tool = os.path.join(
    settings.corosync_qnet_binaries, "corosync-qnetd-tool"
)
__qdevice_certutil = os.path.join(
    settings.corosync_qdevice_binaries, "corosync-qdevice-net-certutil"
)
__nss_certificate_db_files = (
    # old NSS DB format
    "cert8.db",
    "key3.db",
    "secmod.db",
    # new NSS DB format
    "cert9.db",
    "key4.db",
    "pkcs11.txt",
)


class QnetdNotRunningException(Exception):
    pass


def set_up_client_certificates(
    runner,
    reporter,
    communicator_factory,
    qnetd_target,
    cluster_name,
    cluster_nodes_target_list,
    skip_offline_nodes,
    allow_skip_offline=True,
):
    """
    setup cluster nodes for using qdevice model net
    CommandRunner runner -- command runner instance
    ReportProcessor reporter -- report processor instance
    NodeCommunicatorFactory communicator_factory -- communicator facto. instance
    Target qnetd_target -- qdevice provider (qnetd host)
    string cluster_name -- name of the cluster to which qdevice is being added
    list cluster_nodes_target_list -- list of cluster nodes targets
    bool skip_offline_nodes -- continue even if not all nodes are accessible
    bool allow_skip_offline -- enables forcing errors by skip_offline_nodes
    """
    reporter.report(
        ReportItem.info(
            reports.messages.QdeviceCertificateDistributionStarted()
        )
    )
    # get qnetd CA certificate
    com_cmd = qdevice_net_com.GetCaCert(reporter)
    com_cmd.set_targets([qnetd_target])
    qnetd_ca_cert = run_and_raise(
        communicator_factory.get_communicator(), com_cmd
    )[0][1]
    # init certificate storage on all nodes
    com_cmd = qdevice_net_com.ClientSetup(
        reporter, qnetd_ca_cert, skip_offline_nodes, allow_skip_offline
    )
    com_cmd.set_targets(cluster_nodes_target_list)
    run_and_raise(communicator_factory.get_communicator(), com_cmd)
    # create client certificate request
    cert_request = client_generate_certificate_request(runner, cluster_name)
    # sign the request on qnetd host
    com_cmd = qdevice_net_com.SignCertificate(reporter)
    com_cmd.add_request(qnetd_target, cert_request, cluster_name)
    signed_certificate = run_and_raise(
        communicator_factory.get_communicator(), com_cmd
    )[0][1]
    # transform the signed certificate to pk12 format which can sent to nodes
    pk12 = client_cert_request_to_pk12(runner, signed_certificate)
    # distribute final certificate to nodes
    com_cmd = qdevice_net_com.ClientImportCertificateAndKey(
        reporter, pk12, skip_offline_nodes, allow_skip_offline
    )
    com_cmd.set_targets(cluster_nodes_target_list)
    run_and_raise(communicator_factory.get_communicator(), com_cmd)


def qdevice_setup(runner):
    """
    initialize qdevice on local host
    """
    if qdevice_initialized():
        raise LibraryError(
            ReportItem.error(
                reports.messages.QdeviceAlreadyInitialized(__model)
            )
        )

    stdout, stderr, retval = runner.run([__qnetd_certutil, "-i"])
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.QdeviceInitializationError(
                    __model,
                    join_multilines([stderr, stdout]),
                )
            )
        )


def qdevice_initialized():
    """
    check if qdevice server certificate database has been initialized
    """
    return _nss_certificate_db_initialized(
        settings.corosync_qdevice_net_server_certs_dir
    )


def qdevice_destroy():
    """
    delete qdevice configuration on local host
    """
    try:
        if qdevice_initialized():
            shutil.rmtree(settings.corosync_qdevice_net_server_certs_dir)
    except EnvironmentError as e:
        raise LibraryError(
            ReportItem.error(
                reports.messages.QdeviceDestroyError(__model, e.strerror)
            )
        ) from e


def qdevice_status_generic_text(runner, verbose=False):
    """
    get qdevice runtime status in plain text
    bool verbose get more detailed output
    """
    args = ["-s"]
    if verbose:
        args.append("-v")
    stdout, stderr, retval = _qdevice_run_tool(runner, args)
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.QdeviceGetStatusError(
                    __model,
                    join_multilines([stderr, stdout]),
                )
            )
        )
    return stdout


def qdevice_status_cluster_text(runner, cluster=None, verbose=False):
    """
    get qdevice runtime status in plain text
    bool verbose get more detailed output
    string cluster show information only about specified cluster
    """
    args = ["-l"]
    if verbose:
        args.append("-v")
    if cluster:
        args.extend(["-c", cluster])
    stdout, stderr, retval = _qdevice_run_tool(runner, args)
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.QdeviceGetStatusError(
                    __model,
                    join_multilines([stderr, stdout]),
                )
            )
        )
    return stdout


def qdevice_connected_clusters(status_cluster_text):
    """
    parse qnetd cluster status listing and return connected clusters' names
    string status_cluster_text output of corosync-qnetd-tool -l
    """
    connected_clusters = []
    regexp = re.compile(r'^Cluster "(?P<cluster>[^"]+)":$')
    for line in status_cluster_text.splitlines():
        match = regexp.search(line)
        if match:
            connected_clusters.append(match.group("cluster"))
    return connected_clusters


def _qdevice_run_tool(runner, args):
    """
    run corosync-qnetd-tool, raise QnetdNotRunningException if qnetd not running
    CommandRunner runner
    iterable args corosync-qnetd-tool arguments
    """
    stdout, stderr, retval = runner.run([__qnetd_tool] + args)
    if retval == 3 and "is qnetd running?" in stderr.lower():
        raise QnetdNotRunningException()
    return stdout, stderr, retval


def qdevice_sign_certificate_request(runner, cert_request, cluster_name):
    """
    sign client certificate request
    cert_request certificate request data
    string cluster_name name of the cluster to which qdevice is being added
    """
    if not qdevice_initialized():
        raise LibraryError(
            ReportItem.error(reports.messages.QdeviceNotInitialized(__model))
        )
    # save the certificate request, corosync tool only works with files
    tmpfile = _store_to_tmpfile(
        cert_request, reports.messages.QdeviceCertificateSignError
    )
    # sign the request
    stdout, stderr, retval = runner.run(
        [__qnetd_certutil, "-s", "-c", tmpfile.name, "-n", cluster_name]
    )
    tmpfile.close()  # temp file is deleted on close
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.QdeviceCertificateSignError(
                    join_multilines([stderr, stdout]),
                )
            )
        )
    # get signed certificate, corosync tool only works with files
    return _get_output_certificate(
        stdout,
        # pylint: disable=unnecessary-lambda
        lambda reason: reports.messages.QdeviceCertificateSignError(reason),
    )


def client_setup(runner, ca_certificate):
    """
    initialize qdevice client on local host
    ca_certificate qnetd CA certificate
    """
    client_destroy()
    # save CA certificate, corosync tool only works with files
    ca_file_path = os.path.join(
        settings.corosync_qdevice_net_client_certs_dir,
        settings.corosync_qdevice_net_client_ca_file_name,
    )
    try:
        if not os.path.exists(ca_file_path):
            os.makedirs(
                settings.corosync_qdevice_net_client_certs_dir, mode=0o700
            )
        with open(ca_file_path, "wb") as ca_file:
            ca_file.write(ca_certificate)
    except EnvironmentError as e:
        raise LibraryError(
            ReportItem.error(
                reports.messages.QdeviceInitializationError(
                    __model,
                    e.strerror,
                )
            )
        ) from e
    # initialize client's certificate storage
    stdout, stderr, retval = runner.run(
        [__qdevice_certutil, "-i", "-c", ca_file_path]
    )
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.QdeviceInitializationError(
                    __model,
                    join_multilines([stderr, stdout]),
                )
            )
        )


def client_initialized():
    """
    check if qdevice net client certificate database has been initialized
    """
    return _nss_certificate_db_initialized(
        settings.corosync_qdevice_net_client_certs_dir
    )


def client_destroy():
    """
    delete qdevice client config files on local host
    """
    try:
        if client_initialized():
            shutil.rmtree(settings.corosync_qdevice_net_client_certs_dir)
    except EnvironmentError as e:
        raise LibraryError(
            ReportItem.error(
                reports.messages.QdeviceDestroyError(__model, e.strerror)
            )
        ) from e


def client_generate_certificate_request(runner, cluster_name):
    """
    create a certificate request which can be signed by qnetd server
    string cluster_name name of the cluster to which qdevice is being added
    """
    if not client_initialized():
        raise LibraryError(
            ReportItem.error(reports.messages.QdeviceNotInitialized(__model))
        )
    stdout, stderr, retval = runner.run(
        [__qdevice_certutil, "-r", "-n", cluster_name]
    )
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.QdeviceInitializationError(
                    __model,
                    join_multilines([stderr, stdout]),
                )
            )
        )
    return _get_output_certificate(
        stdout,
        lambda reason: reports.messages.QdeviceInitializationError(
            __model,
            reason,
        ),
    )


def client_cert_request_to_pk12(runner, cert_request):
    """
    transform signed certificate request to pk12 certificate which can be
    imported to nodes
    cert_request signed certificate request
    """
    if not client_initialized():
        raise LibraryError(
            ReportItem.error(reports.messages.QdeviceNotInitialized(__model))
        )
    # save the signed certificate request, corosync tool only works with files
    tmpfile = _store_to_tmpfile(
        cert_request,
        reports.messages.QdeviceCertificateImportError,
    )
    # transform it
    stdout, stderr, retval = runner.run(
        [__qdevice_certutil, "-M", "-c", tmpfile.name]
    )
    tmpfile.close()  # temp file is deleted on close
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.QdeviceCertificateImportError(
                    join_multilines([stderr, stdout]),
                )
            )
        )
    # get resulting pk12, corosync tool only works with files
    return _get_output_certificate(
        stdout,
        # pylint: disable=unnecessary-lambda
        lambda reason: reports.messages.QdeviceCertificateImportError(reason),
    )


def client_import_certificate_and_key(runner, pk12_certificate):
    """
    import qdevice client certificate to the local node certificate storage
    """
    if not client_initialized():
        raise LibraryError(
            ReportItem.error(reports.messages.QdeviceNotInitialized(__model))
        )
    # save the certificate, corosync tool only works with files
    tmpfile = _store_to_tmpfile(
        pk12_certificate,
        reports.messages.QdeviceCertificateImportError,
    )
    stdout, stderr, retval = runner.run(
        [__qdevice_certutil, "-m", "-c", tmpfile.name]
    )
    tmpfile.close()  # temp file is deleted on close
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.QdeviceCertificateImportError(
                    join_multilines([stderr, stdout]),
                )
            )
        )


def _nss_certificate_db_initialized(cert_db_path):
    for filename in __nss_certificate_db_files:
        if os.path.exists(os.path.join(cert_db_path, filename)):
            return True
    return False


def _store_to_tmpfile(data, report_item_message):
    try:
        return write_tmpfile(data, binary=True)
    except EnvironmentError as e:
        raise LibraryError(
            ReportItem.error(report_item_message(e.strerror))
        ) from e


def _get_output_certificate(cert_tool_output, report_message_func):
    regexp = re.compile(r"^Certificate( request)? stored in (?P<path>.+)$")
    filename = None
    for line in cert_tool_output.splitlines():
        match = regexp.search(line)
        if match:
            filename = match.group("path")
    if not filename:
        raise LibraryError(
            ReportItem.error(report_message_func(cert_tool_output))
        )
    try:
        with open(filename, "rb") as cert_file:
            return cert_file.read()
    except EnvironmentError as e:
        raise LibraryError(
            ReportItem.error(
                report_message_func(
                    "{path}: {error}".format(path=filename, error=e.strerror)
                )
            )
        ) from e



from enum import auto
from typing import (
    MutableSequence,
    Set,
    Tuple,
    Union,
)

from pcs.common.tools import AutoNameEnum

StringSequence = Union[MutableSequence[str], Tuple[str, ...]]
StringCollection = Union[StringSequence, Set[str]]


class CibRuleExpressionType(AutoNameEnum):
    RULE = auto()
    EXPRESSION = auto()  # node attribute expression, named 'expression' in CIB
    DATE_EXPRESSION = auto()
    OP_EXPRESSION = auto()
    RSC_EXPRESSION = auto()


class CibRuleInEffectStatus(AutoNameEnum):
    NOT_YET_IN_EFFECT = auto()
    IN_EFFECT = auto()
    EXPIRED = auto()
    UNKNOWN = auto()


class ResourceRelationType(AutoNameEnum):
    ORDER = auto()
    ORDER_SET = auto()
    INNER_RESOURCES = auto()
    OUTER_RESOURCE = auto()
    RSC_PRIMITIVE = auto()
    RSC_CLONE = auto()
    RSC_GROUP = auto()
    RSC_BUNDLE = auto()
    RSC_UNKNOWN = auto()


class DrRole(AutoNameEnum):
    PRIMARY = auto()
    RECOVERY = auto()


class UnknownCorosyncTransportTypeException(Exception):
    def __init__(self, transport: str) -> None:
        super().__init__()
        self.transport = transport


class CorosyncTransportType(AutoNameEnum):
    UDP = auto()
    UDPU = auto()
    KNET = auto()

    @classmethod
    def from_str(cls, transport: str) -> "CorosyncTransportType":
        try:
            return cls(transport.upper())
        except ValueError:
            raise UnknownCorosyncTransportTypeException(transport) from None



import logging
import os
import re
import tempfile
from functools import lru_cache
from unittest import (
    mock,
    skipUnless,
)

from lxml import etree

from pcs import settings
from pcs.cli.common.parse_args import InputModifiers
from pcs.common import str_tools
from pcs.lib.external import CommandRunner

from pcs_test import TEST_ROOT
from pcs_test import settings as tests_settings
from pcs_test.tools.custom_mock import MockLibraryReportProcessor

runner = CommandRunner(
    mock.MagicMock(logging.Logger), MockLibraryReportProcessor(), os.environ
)


class ParametrizedTestMetaClass(type):
    """
    Example:
        class GeneralTest(TestCase):
            attr = None
            def _test_1(self):
                self.assertIn(self.attr, [1, 2])

            def _test_2(self):
                self.assertNotIn(self.attr, [0, 3, 4, 5])

        class Test1(GeneralTest, metaclass=ParametrizedTestMetaClass):
            attr = 1

        class Test2(GeneralTest, metaclass=ParametrizedTestMetaClass):
            attr = 2

        class Test3(GeneralTest, metaclass=ParametrizedTestMetaClass):
            # This should fail
            attr = 3
    """

    def __init__(cls, classname, bases, class_dict):
        for attr_name in dir(cls):
            attr = getattr(cls, attr_name)
            if attr_name.startswith("_test") and hasattr(attr, "__call__"):
                setattr(cls, attr_name[1:], attr)

        super().__init__(classname, bases, class_dict)


def dict_to_modifiers(options):
    def _convert_val(val):
        if val is True:
            return ""
        return val

    return InputModifiers(
        {
            f"--{opt}": _convert_val(val)
            for opt, val in options.items()
            if val is not False
        }
    )


def get_test_resource(name):
    """Return full path to a test resource file specified by name"""
    return os.path.join(TEST_ROOT, "resources", name)


def get_tmp_dir(name=None):
    """Create a temp directory with a unique name in our test dir"""
    # pylint: disable=consider-using-with
    tmp_dir = get_test_resource("temp")
    os.makedirs(tmp_dir, exist_ok=True)
    return tempfile.TemporaryDirectory(
        suffix=".tmp",  # for .gitignore
        prefix=(f"{name}." if name else None),
        dir=tmp_dir,
    )


def get_tmp_file(name=None, mode="w+"):
    """Create a temp file with a unique name in our test dir"""
    # pylint: disable=consider-using-with
    tmp_dir = get_test_resource("temp")
    os.makedirs(tmp_dir, exist_ok=True)
    return tempfile.NamedTemporaryFile(
        mode=mode,
        suffix=".tmp",  # for .gitignore
        prefix=(f"{name}." if name else None),
        dir=tmp_dir,
    )


def write_data_to_tmpfile(data, tmp_file):
    tmp_file.seek(0)
    tmp_file.truncate()
    tmp_file.write(data)
    tmp_file.flush()
    tmp_file.seek(0)


def write_file_to_tmpfile(source_file_path, tmp_file):
    with open(source_file_path) as source:
        write_data_to_tmpfile(source.read(), tmp_file)


def read_test_resource(name):
    with open(get_test_resource(name)) as a_file:
        return a_file.read()


def cmp3(a, b):
    # pylint: disable=invalid-name

    # python3 doesn't have the cmp function, this is an official workaround
    # https://docs.python.org/3.0/whatsnew/3.0.html#ordering-comparisons
    return (a > b) - (a < b)


def compare_version(a, b):
    # pylint: disable=invalid-name
    if a[0] == b[0]:
        if a[1] == b[1]:
            return cmp3(a[2], b[2])
        return cmp3(a[1], b[1])
    return cmp3(a[0], b[0])


def is_minimum_pacemaker_version(major, minor, rev):
    return _is_version_sufficient(
        _get_current_pacemaker_version(), (major, minor, rev)
    )


@lru_cache()
def _get_current_pacemaker_version():
    output, dummy_stderr, dummy_retval = runner.run(
        [
            os.path.join(settings.pacemaker_binaries, "crm_mon"),
            "--version",
        ]
    )
    pacemaker_version = output.split("\n", maxsplit=1)[0]
    regexp = re.compile(r"Pacemaker (\d+)\.(\d+)\.(\d+)")
    match = regexp.match(pacemaker_version)
    major = int(match.group(1))
    minor = int(match.group(2))
    rev = int(match.group(3))
    return major, minor, rev


@lru_cache()
def _get_current_cib_schema_version():
    regexp = re.compile(r"pacemaker-((\d+)\.(\d+))")
    all_versions = set()
    xml = etree.parse(tests_settings.pacemaker_version_rng).getroot()
    for value_el in xml.xpath(
        ".//x:attribute[@name='validate-with']//x:value",
        namespaces={"x": "http://relaxng.org/ns/structure/1.0"},
    ):
        match = re.match(regexp, value_el.text)
        if match:
            all_versions.add((int(match.group(2)), int(match.group(3))))
    return sorted(all_versions)[-1]


def _is_minimum_cib_schema_version(cmajor, cminor, crev):
    major, minor = _get_current_cib_schema_version()
    return compare_version((major, minor, 0), (cmajor, cminor, crev)) > -1


def _is_version_sufficient(current_version, minimal_version):
    return compare_version(current_version, minimal_version) > -1


def _format_version(version_tuple):
    return ".".join([str(x) for x in version_tuple])


def is_minimum_pacemaker_features(cmajor, cminor, crev):
    major, minor, rev = _get_current_pacemaker_features()[0]
    return compare_version((major, minor, rev), (cmajor, cminor, crev)) > -1


def _has_pacemaker_features(requested_features):
    requested = frozenset(requested_features)
    actual = _get_current_pacemaker_features()[1]
    return actual >= requested


@lru_cache()
def _get_current_pacemaker_features():
    output, dummy_stderr, dummy_retval = runner.run(
        [
            os.path.join(settings.pacemaker_binaries, "pacemakerd"),
            "--features",
        ]
    )
    features_string = output.split("\n")[1]
    regexp = re.compile(r"Supporting v(\d+)\.(\d+)\.(\d+):\s*(.*)")
    match = regexp.search(features_string)
    major = int(match.group(1))
    minor = int(match.group(2))
    rev = int(match.group(3))
    features_list = frozenset(match.group(4).split())
    return (major, minor, rev), features_list


@lru_cache()
def is_pacemaker_21_without_20_compatibility():
    return is_minimum_pacemaker_version(
        2, 1, 0
    ) and not _has_pacemaker_features(["compat-2.0"])


def skip_unless_pacemaker_version(version_tuple, feature):
    current_version = _get_current_pacemaker_version()
    return skipUnless(
        _is_version_sufficient(current_version, version_tuple),
        (
            "Pacemaker version is too old (current: {current_version},"
            " must be >= {minimal_version}) to test {feature}"
        ).format(
            current_version=_format_version(current_version),
            minimal_version=_format_version(version_tuple),
            feature=feature,
        ),
    )


def skip_unless_pacemaker_features(version_tuple, feature):
    return skipUnless(
        is_minimum_pacemaker_features(*version_tuple),
        (
            "Pacemaker must support feature set version {version} to test "
            "{feature}"
        ).format(version=_format_version(version_tuple), feature=feature),
    )


def skip_unless_cib_schema_version(version_tuple, feature):
    current_version = _get_current_cib_schema_version()
    return skipUnless(
        _is_minimum_cib_schema_version(*version_tuple),
        (
            "Pacemaker supported CIB schema version is too low (current: "
            "{current_version}, must be >= {minimal_version}) to test {feature}"
        ).format(
            current_version=_format_version(current_version),
            minimal_version=_format_version(version_tuple),
            feature=feature,
        ),
    )


def skip_unless_crm_rule():
    return skip_unless_pacemaker_version(
        (2, 0, 2), "listing of constraints that might be expired"
    )


def skip_unless_pacemaker_supports_rsc_and_op_rules():
    return skip_unless_cib_schema_version(
        (3, 4, 0), "rsc_expression and op_expression elements in rule elements"
    )


def skip_unless_pacemaker_supports_op_onfail_demote():
    return skip_unless_cib_schema_version(
        (3, 4, 0), "resource operations with 'on-fail' option set to 'demote'"
    )


def skip_unless_root():
    return skipUnless(os.getuid() == 0, "Root user required")


@lru_cache()
def _is_booth_resource_agent_installed():
    output, dummy_stderr, dummy_retval = runner.run(
        [
            os.path.join(settings.pacemaker_binaries, "crm_resource"),
            "--list-agents",
            "ocf:pacemaker",
        ]
    )
    return "booth-site" in output


def skip_unless_booth_resource_agent_installed():
    return skipUnless(
        _is_booth_resource_agent_installed(),
        "test requires resource agent ocf:pacemaker:booth-site"
        " which is not installed",
    )


def create_patcher(target_prefix_or_module):
    """
    Return function for patching tests with preconfigured target prefix
    string|module target_prefix_or_module could be:
        * a prefix for patched names. Typically tested module:
            "pcs.lib.commands.booth"
        * a (imported) module: pcs.lib.cib
        Between prefix and target is "." (dot)
    """
    prefix = target_prefix_or_module
    if not isinstance(target_prefix_or_module, str):
        prefix = target_prefix_or_module.__name__

    def patch(target, *args, **kwargs):
        return mock.patch("{0}.{1}".format(prefix, target), *args, **kwargs)

    return patch


def outdent(text):
    return "\n".join(str_tools.outdent(text.splitlines()))


def create_setup_patch_mixin(module_specification_or_patcher):
    """
    Configure and return SetupPatchMixin

    SetupPatchMixin add method 'setup_patch' to a test case.

    Method setup_patch takes name that should be patched in destination module
    (see module_specification_or_patcher). Method provide cleanup after test.
    It is expected to be used in 'setUp' method but should work inside test as
    well.

    string|callable module_specification_or_patcher can be
       * callable patcher created via create_patcher:
         create_patcher("pcs.lib.cib")
       * name of module: "pcs.lib.cib"
       * (imported) module: pcs.lib.cib
         Note that this must be not a callable (can be done via
         sys.modules[__name__] = something_callable. If is a callable use name
         of the module instead.
    """
    if callable(module_specification_or_patcher):
        patch_module = module_specification_or_patcher
    else:
        patch_module = create_patcher(module_specification_or_patcher)

    class SetupPatchMixin:
        def setup_patch(self, target_suffix, *args, **kwargs):
            patcher = patch_module(target_suffix, *args, **kwargs)
            self.addCleanup(patcher.stop)
            return patcher.start()

    return SetupPatchMixin



from textwrap import dedent
from unittest import TestCase

from pcs.common.corosync_conf import (
    CorosyncConfDto,
    CorosyncNodeAddressDto,
    CorosyncNodeDto,
    CorosyncQuorumDeviceSettingsDto,
)
from pcs.common.reports import codes as report_codes
from pcs.common.types import CorosyncTransportType
from pcs.lib.commands import cluster

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

from .common import fixture_totem


class GetCorosyncConfStruct(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_unsupported_corosync_transport(self):
        self.config.corosync_conf.load_content(
            fixture_totem(transport_type="unknown")
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.get_corosync_conf_struct(self.env_assist.get_env()),
            [
                fixture.error(
                    report_codes.COROSYNC_CONFIG_UNSUPPORTED_TRANSPORT,
                    actual_transport="unknown",
                    supported_transport_types=["knet", "udp", "udpu"],
                ),
            ],
            expected_in_processor=False,
        )

    def test_empty_corosync_conf(self):
        self.config.corosync_conf.load_content("")
        self.assertEqual(
            CorosyncConfDto(
                cluster_name="",
                cluster_uuid=None,
                transport=CorosyncTransportType.KNET,
                totem_options={},
                transport_options={},
                compression_options={},
                crypto_options={},
                nodes=[],
                links_options={},
                quorum_options={},
                quorum_device=None,
            ),
            cluster.get_corosync_conf_struct(self.env_assist.get_env()),
        )

    def test_corosync_conf_with_uuid(self):
        self.config.corosync_conf.load_content(
            dedent(
                """\
                totem {
                    version: 2
                    cluster_name: HACluster
                    cluster_uuid: uuid
                    transport: knet
                    crypto_cipher: aes256
                    crypto_hash: sha256
                }
                
                nodelist {
                    node {
                        ring0_addr: node1-addr
                        name: node1
                        nodeid: 1
                    }

                    node {
                        ring0_addr: node2-addr
                        name: node2
                        nodeid: 2
                    }
                }
                
                quorum {
                    provider: corosync_votequorum
                    two_node: 1
                }
                """
            )
        )
        self.assertEqual(
            CorosyncConfDto(
                cluster_name="HACluster",
                cluster_uuid="uuid",
                transport=CorosyncTransportType.KNET,
                totem_options={},
                transport_options={},
                compression_options={},
                crypto_options={"cipher": "aes256", "hash": "sha256"},
                nodes=[
                    CorosyncNodeDto(
                        name="node1",
                        nodeid="1",
                        addrs=[
                            CorosyncNodeAddressDto(
                                addr="node1-addr",
                                link="0",
                                type="FQDN",
                            ),
                        ],
                    ),
                    CorosyncNodeDto(
                        name="node2",
                        nodeid="2",
                        addrs=[
                            CorosyncNodeAddressDto(
                                addr="node2-addr",
                                link="0",
                                type="FQDN",
                            ),
                        ],
                    ),
                ],
                links_options={},
                quorum_options={},
                quorum_device=None,
            ),
            cluster.get_corosync_conf_struct(self.env_assist.get_env()),
        )

    def test_corosync_conf_with_qdevice(self):
        self.config.corosync_conf.load_content(
            dedent(
                """\
                totem {
                    version: 2
                    cluster_name: HACluster
                    transport: knet
                    ip_version: ipv4-6
                    link_mode: passive
                    knet_compression_level: 5
                    knet_compression_model: zlib
                    knet_compression_threshold: 100
                    crypto_cipher: aes256
                    crypto_hash: sha256
                    consensus: 3600
                    join: 50
                    token: 3000

                    interface {
                        linknumber: 0
                        knet_link_priority: 100
                        knet_ping_interval: 750
                        knet_ping_timeout: 1500
                        knet_transport: udp
                    }

                    interface {
                        linknumber: 1
                        knet_link_priority: 200
                        knet_ping_interval: 750
                        knet_ping_timeout: 1500
                        knet_transport: sctp
                    }
                }

                nodelist {
                    node {
                        ring0_addr: node1-addr
                        ring1_addr: 10.0.0.1
                        name: node1
                        nodeid: 1
                    }

                    node {
                        ring0_addr: node2-addr
                        ring1_addr: 10.0.0.2
                        name: node2
                        nodeid: 2
                    }
                }

                quorum {
                    provider: corosync_votequorum
                    two_node: 1
                    wait_for_all: 1
                    device {
                        model: net
                        sync_timeout: 5000
                        timeout: 5000
                        net {
                            algorithm: ffsplit
                            host: node-qdevice
                        }
                        heuristics {
                            mode: on
                            exec_ping: /usr/bin/ping -c 1 127.0.0.1
                        }
                    }
                }

                logging {
                    to_logfile: yes
                    logfile: /var/log/cluster/corosync.log
                    to_syslog: yes
                    timestamp: on
                }
                """
            )
        )
        self.assertEqual(
            CorosyncConfDto(
                cluster_name="HACluster",
                cluster_uuid=None,
                transport=CorosyncTransportType.KNET,
                totem_options={
                    "consensus": "3600",
                    "join": "50",
                    "token": "3000",
                },
                transport_options={
                    "ip_version": "ipv4-6",
                    "link_mode": "passive",
                },
                crypto_options={"cipher": "aes256", "hash": "sha256"},
                compression_options={
                    "level": "5",
                    "model": "zlib",
                    "threshold": "100",
                },
                nodes=[
                    CorosyncNodeDto(
                        name="node1",
                        nodeid="1",
                        addrs=[
                            CorosyncNodeAddressDto(
                                addr="node1-addr",
                                link="0",
                                type="FQDN",
                            ),
                            CorosyncNodeAddressDto(
                                addr="10.0.0.1",
                                link="1",
                                type="IPv4",
                            ),
                        ],
                    ),
                    CorosyncNodeDto(
                        name="node2",
                        nodeid="2",
                        addrs=[
                            CorosyncNodeAddressDto(
                                addr="node2-addr",
                                link="0",
                                type="FQDN",
                            ),
                            CorosyncNodeAddressDto(
                                addr="10.0.0.2",
                                link="1",
                                type="IPv4",
                            ),
                        ],
                    ),
                ],
                links_options={
                    "0": {
                        "linknumber": "0",
                        "link_priority": "100",
                        "ping_interval": "750",
                        "ping_timeout": "1500",
                        "transport": "udp",
                    },
                    "1": {
                        "linknumber": "1",
                        "link_priority": "200",
                        "ping_interval": "750",
                        "ping_timeout": "1500",
                        "transport": "sctp",
                    },
                },
                quorum_options={"wait_for_all": "1"},
                quorum_device=CorosyncQuorumDeviceSettingsDto(
                    model="net",
                    model_options={
                        "algorithm": "ffsplit",
                        "host": "node-qdevice",
                    },
                    generic_options={"sync_timeout": "5000", "timeout": "5000"},
                    heuristics_options={
                        "mode": "on",
                        "exec_ping": "/usr/bin/ping -c 1 127.0.0.1",
                    },
                ),
            ),
            cluster.get_corosync_conf_struct(self.env_assist.get_env()),
        )



import base64

from pcs import settings
from pcs.common import file_type_codes
from pcs.common.reports import codes as report_codes

from pcs_test.tools import fixture

OFFLINE_ERROR_MSG = "Could not resolve host"
FAIL_HTTP_KWARGS = dict(
    output="",
    was_connected=False,
    errno="6",
    error_msg=OFFLINE_ERROR_MSG,
)


class EnvConfigMixin:
    def __init__(self, call_collection, wrap_helper, config):
        del wrap_helper, call_collection
        self.config = config

    def distribute_authkey(
        self, communication_list, pcmk_authkey_content, result=None, **kwargs
    ):
        if kwargs.get("was_connected", True):
            result = (
                result
                if result is not None
                else {
                    "code": "written",
                    "message": "",
                }
            )

            kwargs["results"] = {"pacemaker_remote authkey": result}
        elif result is not None:
            raise AssertionError(
                "Keyword 'result' makes no sense with 'was_connected=False'"
            )
        self.config.http.put_file(
            communication_list=communication_list,
            files={
                "pacemaker_remote authkey": {
                    "type": "pcmk_remote_authkey",
                    "data": base64.b64encode(pcmk_authkey_content).decode(
                        "utf-8"
                    ),
                    "rewrite_existing": True,
                }
            },
            **kwargs,
        )

    def get_host_info(self, label, dest_list, output=None):
        if output is None:
            output = dict(
                services=dict(
                    pacemaker_remote=dict(
                        installed=True, enabled=False, running=False
                    ),
                ),
                cluster_configuration_exists=False,
            )
        self.config.http.host.get_host_info(
            communication_list=[
                {
                    "label": label,
                    "dest_list": dest_list,
                },
            ],
            output_data=output,
        )

    def authkey_exists(self, return_value):
        self.config.raw_file.exists(
            file_type_codes.PACEMAKER_AUTHKEY,
            settings.pacemaker_authkey_file,
            exists=return_value,
        )

    def open_authkey(self, pcmk_authkey_content="", fail=False):
        self.config.raw_file.read(
            file_type_codes.PACEMAKER_AUTHKEY,
            settings.pacemaker_authkey_file,
            content=(pcmk_authkey_content if not fail else None),
            exception_msg=("open failed" if fail else None),
        )

    def push_existing_authkey_to_remote(
        self, host_name, dest_list, distribution_result=None
    ):
        pcmk_authkey_content = b"password"
        (
            self.config.local.authkey_exists(return_value=True)
            .local.open_authkey(pcmk_authkey_content)
            .local.distribute_authkey(
                communication_list=[dict(label=host_name, dest_list=dest_list)],
                pcmk_authkey_content=pcmk_authkey_content,
                result=distribution_result,
            )
        )

    def run_pacemaker_remote(self, label, dest_list, result=None, **kwargs):
        if kwargs.get("was_connected", True):
            result = (
                result
                if result is not None
                else {
                    "code": "success",
                    "message": "",
                }
            )

            kwargs["results"] = {
                "pacemaker_remote enable": result,
                "pacemaker_remote start": result,
            }
        elif result is not None:
            raise AssertionError(
                "Keyword 'result' makes no sense with 'was_connected=False'"
            )

        self.config.http.manage_services(
            communication_list=[dict(label=label, dest_list=dest_list)],
            action_map={
                "pacemaker_remote enable": {
                    "type": "service_command",
                    "service": "pacemaker_remote",
                    "command": "enable",
                },
                "pacemaker_remote start": {
                    "type": "service_command",
                    "service": "pacemaker_remote",
                    "command": "start",
                },
            },
            **kwargs,
        )


REPORTS = (
    fixture.ReportStore()
    .info(
        "authkey_distribution_started",
        report_codes.FILES_DISTRIBUTION_STARTED,
        file_list=["pacemaker authkey"],
    )
    .info(
        "authkey_distribution_success",
        report_codes.FILE_DISTRIBUTION_SUCCESS,
        file_description="pacemaker authkey",
    )
    .info(
        "pcmk_remote_start_enable_started",
        report_codes.SERVICE_COMMANDS_ON_NODES_STARTED,
        action_list=[
            "pacemaker_remote start",
            "pacemaker_remote enable",
        ],
    )
    .info(
        "pcmk_remote_enable_success",
        report_codes.SERVICE_COMMAND_ON_NODE_SUCCESS,
        service_command_description="pacemaker_remote enable",
    )
    .info(
        "pcmk_remote_start_success",
        report_codes.SERVICE_COMMAND_ON_NODE_SUCCESS,
        service_command_description="pacemaker_remote start",
    )
)

EXTRA_REPORTS = (
    fixture.ReportStore()
    .error(
        "manage_services_connection_failed",
        report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
        command="remote/manage_services",
        reason=OFFLINE_ERROR_MSG,
        force_code=report_codes.SKIP_OFFLINE_NODES,
    )
    .as_warn(
        "manage_services_connection_failed",
        "manage_services_connection_failed_warn",
    )
    .copy(
        "manage_services_connection_failed",
        "put_file_connection_failed",
        command="remote/put_file",
    )
    .as_warn(
        "put_file_connection_failed",
        "put_file_connection_failed_warn",
    )
    .error(
        "pcmk_remote_enable_failed",
        report_codes.SERVICE_COMMAND_ON_NODE_ERROR,
        reason="Operation failed.",
        service_command_description="pacemaker_remote enable",
        force_code=report_codes.FORCE,
    )
    .as_warn("pcmk_remote_enable_failed", "pcmk_remote_enable_failed_warn")
    .copy(
        "pcmk_remote_enable_failed",
        "pcmk_remote_start_failed",
        service_command_description="pacemaker_remote start",
    )
    .as_warn("pcmk_remote_start_failed", "pcmk_remote_start_failed_warn")
    .error(
        "authkey_distribution_failed",
        report_codes.FILE_DISTRIBUTION_ERROR,
        reason="File already exists",
        file_description="pacemaker authkey",
        force_code=report_codes.FORCE,
    )
    .as_warn("authkey_distribution_failed", "authkey_distribution_failed_warn")
)


def fixture_reports_not_live_cib(node_name):
    return [
        fixture.info(
            report_codes.COROSYNC_NODE_CONFLICT_CHECK_SKIPPED,
            reason_type="not_live_cib",
        ),
        fixture.info(
            report_codes.FILES_DISTRIBUTION_SKIPPED,
            reason_type="not_live_cib",
            file_list=["pacemaker authkey"],
            node_list=[node_name],
        ),
        fixture.info(
            report_codes.SERVICE_COMMANDS_ON_NODES_SKIPPED,
            reason_type="not_live_cib",
            action_list=[
                "pacemaker_remote start",
                "pacemaker_remote enable",
            ],
            node_list=[node_name],
        ),
    ]


def fixture_reports_new_node_unreachable(node_name, omitting=False):
    if omitting:
        report = [
            fixture.warn(
                report_codes.OMITTING_NODE,
                node=node_name,
            ),
        ]
    else:
        report = [
            fixture.warn(
                report_codes.HOST_NOT_FOUND,
                host_list=[node_name],
            ),
        ]
    return report + [
        fixture.info(
            report_codes.FILES_DISTRIBUTION_SKIPPED,
            reason_type="unreachable",
            file_list=["pacemaker authkey"],
            node_list=[node_name],
        ),
        fixture.info(
            report_codes.SERVICE_COMMANDS_ON_NODES_SKIPPED,
            reason_type="unreachable",
            action_list=["pacemaker_remote start", "pacemaker_remote enable"],
            node_list=[node_name],
        ),
    ]



from textwrap import dedent
from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import (
    get_tmp_file,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner

from .common import fixture_corosync_conf_minimal


class UpdateLocal(AssertPcsMixin, TestCase):
    def setUp(self):
        self.corosync_conf_file = get_tmp_file(
            "tier1_cluster_config_update_corosync.conf"
        )
        self.pcs_runner = PcsRunner(
            cib_file=None,
            corosync_conf_opt=self.corosync_conf_file.name,
        )

    def tearDown(self):
        self.corosync_conf_file.close()

    def test_minimal(self):
        write_data_to_tmpfile(
            fixture_corosync_conf_minimal(no_cluster_uuid=True),
            self.corosync_conf_file,
        )
        self.assert_pcs_success(
            "cluster config update transport ip_version=ipv4 totem token=12".split()
        )
        self.assertEqual(
            self.corosync_conf_file.read(),
            dedent(
                """\
                totem {
                    version: 2
                    cluster_name: cluster_name
                    transport: knet
                    ip_version: ipv4
                    crypto_cipher: aes256
                    crypto_hash: sha256
                    token: 12
                }

                nodelist {
                    node {
                        ring0_addr: node1_addr
                        name: node1
                        nodeid: 1
                    }

                    node {
                        ring0_addr: node2_addr
                        name: node2
                        nodeid: 2
                    }
                }

                quorum {
                    provider: corosync_votequorum
                    two_node: 1
                }

                logging {
                    to_logfile: yes
                    logfile: /var/log/cluster/corosync.log
                    to_syslog: yes
                    timestamp: on
                }
                """
            ),
        )

    def test_multiple_options(self):
        write_data_to_tmpfile(
            fixture_corosync_conf_minimal(), self.corosync_conf_file
        )
        self.assert_pcs_success(
            (
                "cluster config update "
                "transport ip_version= link_mode=passive "
                "compression level=2 model=zlib threshold=10 "
                "crypto cipher=aes256 hash=sha512 model=openssl "
                "totem consensus=0 downcheck=1 token=12"
            ).split()
        )
        self.assertEqual(
            self.corosync_conf_file.read(),
            dedent(
                """\
                totem {
                    version: 2
                    cluster_name: cluster_name
                    cluster_uuid: cluster_uuid
                    transport: knet
                    crypto_cipher: aes256
                    crypto_hash: sha512
                    consensus: 0
                    downcheck: 1
                    token: 12
                    link_mode: passive
                    knet_compression_level: 2
                    knet_compression_model: zlib
                    knet_compression_threshold: 10
                    crypto_model: openssl
                }

                nodelist {
                    node {
                        ring0_addr: node1_addr
                        name: node1
                        nodeid: 1
                    }

                    node {
                        ring0_addr: node2_addr
                        name: node2
                        nodeid: 2
                    }
                }

                quorum {
                    provider: corosync_votequorum
                    two_node: 1
                }

                logging {
                    to_logfile: yes
                    logfile: /var/log/cluster/corosync.log
                    to_syslog: yes
                    timestamp: on
                }
                """
            ),
        )

    def test_file_does_not_exist(self):
        file_name = self.corosync_conf_file.name + ".non-existing"
        self.pcs_runner.corosync_conf_opt = file_name
        self.assert_pcs_fail(
            "cluster config update transport ip_version=ipv4 totem token=12".split(),
            stdout_full=(
                f"Error: Unable to read Corosync configuration '{file_name}': "
                f"No such file or directory: '{file_name}'\n"
            ),
        )

    def test_file_parse_error(self):
        write_data_to_tmpfile(
            "this is not\na valid corosync.conf file\n", self.corosync_conf_file
        )
        self.assert_pcs_fail(
            "cluster config update transport ip_version=ipv4 totem token=12".split(),
            stdout_full=(
                "Error: Unable to parse corosync config: a line is not opening "
                "or closing a section or key: value\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )

    def test_validation_errors(self):
        write_data_to_tmpfile(
            fixture_corosync_conf_minimal(), self.corosync_conf_file
        )
        self.assert_pcs_fail(
            (
                "cluster config update "
                "transport ip_version=ipvx link_mode=passive "
                "compression level=2 model=zlib threshold=NaN "
                "crypto hash= model=openssl "
                "totem consensus=0 down_check=1 token=12"
            ).split(),
            stdout_full=(
                "Error: invalid totem option 'down_check', allowed options "
                "are: 'block_unlisted_ips', 'consensus', 'downcheck', 'fail_recv_const', "
                "'heartbeat_failures_allowed', 'hold', 'join', 'max_messages', "
                "'max_network_delay', 'merge', 'miss_count_const', "
                "'send_join', 'seqno_unchanged_const', 'token', "
                "'token_coefficient', 'token_retransmit', "
                "'token_retransmits_before_loss_const', 'window_size'\n"
                "Error: 'ipvx' is not a valid ip_version value, use 'ipv4', "
                "'ipv4-6', 'ipv6', 'ipv6-4'\n"
                "Error: 'NaN' is not a valid threshold value, use a "
                "non-negative integer\n"
                "Error: If crypto option 'cipher' is enabled, crypto option "
                "'hash' must be enabled as well\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )
        self.assertEqual(
            self.corosync_conf_file.read(),
            fixture_corosync_conf_minimal(),
        )



from unittest import TestCase

from pcs.cli.common.completion import (
    _find_suggestions,
    _split_words,
    has_applicable_environment,
    make_suggestions,
)

tree = {
    "resource": {
        "op": {
            "add": {},
            "defaults": {},
            "remove": {},
        },
        "clone": {},
    },
    "cluster": {
        "auth": {},
        "cib": {},
    },
}


class SuggestionTest(TestCase):
    def test_suggest_nothing_when_cursor_on_first_word(self):
        self.assertEqual([], _find_suggestions(tree, ["pcs"], 0))
        self.assertEqual([], _find_suggestions(tree, ["pcs", "resource"], 0))

    def test_suggest_nothing_when_cursor_position_out_of_range(self):
        self.assertEqual([], _find_suggestions(tree, ["pcs", "resource"], 3))

    def test_suggest_when_last_word_not_started(self):
        self.assertEqual(
            ["clone", "op"], _find_suggestions(tree, ["pcs", "resource"], 2)
        )

    def test_suggest_when_last_word_started(self):
        self.assertEqual(
            ["clone"], _find_suggestions(tree, ["pcs", "resource", "c"], 2)
        )

    def test_suggest_when_cursor_on_word_amid(self):
        self.assertEqual(
            ["clone"],
            _find_suggestions(tree, ["pcs", "resource", "c", "add"], 2),
        )

    def test_suggest_nothing_when_previously_typed_word_not_match(self):
        self.assertEqual(
            [], _find_suggestions(tree, ["pcs", "invalid", "c"], 2)
        )


class HasCompletionEnvironmentTest(TestCase):
    def test_returns_false_if_environment_inapplicable(self):
        inapplicable_environments = [
            {
                "COMP_CWORD": "1",
                "PCS_AUTO_COMPLETE": "1",
            },
            {
                "COMP_WORDS": "pcs resource",
                "PCS_AUTO_COMPLETE": "1",
            },
            {
                "COMP_WORDS": "pcs resource",
                "COMP_CWORD": "1",
            },
            {
                "COMP_WORDS": "pcs resource",
                "COMP_CWORD": "1",
                "PCS_AUTO_COMPLETE": "0",
            },
            {
                "COMP_WORDS": "pcs resource",
                "COMP_CWORD": "1a",
                "PCS_AUTO_COMPLETE": "1",
            },
            {
                "COMP_WORDS": "pcs resource",
                "COMP_CWORD": "1",
                "PCS_AUTO_COMPLETE": "1",
            },
        ]
        for environment in inapplicable_environments:
            self.assertFalse(
                has_applicable_environment(environment),
                "environment evaluated as applicable (should not be): "
                + repr(environment),
            )

    def test_returns_true_if_environment_is_set(self):
        self.assertTrue(
            has_applicable_environment(
                {
                    "COMP_WORDS": "pcs resource",
                    "COMP_CWORD": "1",
                    "COMP_LENGTHS": "3 8",
                    "PCS_AUTO_COMPLETE": "1",
                }
            )
        )


class MakeSuggestionsEnvironment(TestCase):
    def test_raises_for_incomlete_environment(self):
        self.assertRaises(
            EnvironmentError,
            lambda: make_suggestions(
                {
                    "COMP_CWORD": "1",
                    "PCS_AUTO_COMPLETE": "1",
                },
                suggestion_tree=tree,
            ),
        )

    def test_suggest_on_correct_environment(self):
        self.assertEqual(
            "clone\nop",
            make_suggestions(
                {
                    "COMP_WORDS": "pcs resource",
                    "COMP_CWORD": "2",
                    "COMP_LENGTHS": "3 8",
                    "PCS_AUTO_COMPLETE": "1",
                },
                suggestion_tree=tree,
            ),
        )


class SplitWordsTest(TestCase):
    def test_return_word_list_on_compatible_words_and_length(self):
        self.assertEqual(
            ["pcs", "resource", "op", "a"],
            _split_words("pcs resource op a", ["3", "8", "2", "1"]),
        )

    def test_refuse_when_no_int_in_lengths(self):
        self.assertRaises(
            EnvironmentError,
            lambda: _split_words("pcs resource op a", ["3", "8", "2", "A"]),
        )

    def test_refuse_when_lengths_are_too_big(self):
        self.assertRaises(
            EnvironmentError,
            lambda: _split_words("pcs resource op a", ["3", "8", "2", "10"]),
        )

    def test_refuse_when_separator_doesnot_match(self):
        self.assertRaises(
            EnvironmentError,
            lambda: _split_words("pc sresource op a", ["3", "8", "2", "1"]),
        )

    def test_refuse_when_lengths_are_too_small(self):
        self.assertRaises(
            EnvironmentError,
            lambda: _split_words("pcs resource op a ", ["3", "8", "2", "1"]),
        )



from unittest import (
    TestCase,
    mock,
)

from pcs.common import const
from pcs.lib import resource_agent as ra


class GetAdditionalTraceParameters(TestCase):
    def _assert_param_names(self, parameters, expected_names):
        self.assertEqual([param.name for param in parameters], expected_names)

    @staticmethod
    def _fixture_parameter(name):
        return ra.ResourceAgentParameter(
            name,
            shortdesc=None,
            longdesc=None,
            type="string",
            default=None,
            enum_values=None,
            required=False,
            advanced=False,
            deprecated=False,
            deprecated_by=[],
            deprecated_desc=None,
            unique_group=None,
            reloadable=False,
        )

    def test_no_input_params(self):
        self._assert_param_names(
            ra.pcs_transform.get_additional_trace_parameters([]),
            ["trace_ra", "trace_file"],
        )

    def test_return_both(self):
        self._assert_param_names(
            ra.pcs_transform.get_additional_trace_parameters(
                [
                    self._fixture_parameter("param1"),
                    self._fixture_parameter("trace"),
                ]
            ),
            ["trace_ra", "trace_file"],
        )

    def test_return_trace_file(self):
        self._assert_param_names(
            ra.pcs_transform.get_additional_trace_parameters(
                [
                    self._fixture_parameter("param1"),
                    self._fixture_parameter("trace_ra"),
                ]
            ),
            ["trace_file"],
        )

    def test_return_trace_ra(self):
        self._assert_param_names(
            ra.pcs_transform.get_additional_trace_parameters(
                [
                    self._fixture_parameter("param1"),
                    self._fixture_parameter("trace_file"),
                ]
            ),
            ["trace_ra"],
        )

    def test_return_none(self):
        self._assert_param_names(
            ra.pcs_transform.get_additional_trace_parameters(
                [
                    self._fixture_parameter("param1"),
                    self._fixture_parameter("trace_file"),
                    self._fixture_parameter("trace_ra"),
                ]
            ),
            [],
        )


ra_pkg = "pcs.lib.resource_agent.pcs_transform"


@mock.patch(f"{ra_pkg}._metadata_make_stonith_port_parameter_not_required")
@mock.patch(f"{ra_pkg}._metadata_make_stonith_action_parameter_deprecated")
@mock.patch(f"{ra_pkg}._metadata_remove_unwanted_stonith_parameters")
@mock.patch(f"{ra_pkg}._metadata_parameter_extract_advanced_from_desc")
@mock.patch(f"{ra_pkg}._metadata_parameter_join_short_long_desc")
@mock.patch(f"{ra_pkg}._metadata_action_translate_role")
class OcfUnifiedToPcs(TestCase):
    @staticmethod
    def _fixture_metadata(name):
        return ra.ResourceAgentMetadata(
            name,
            agent_exists=True,
            ocf_version=ra.const.OCF_1_0,
            shortdesc=None,
            longdesc=None,
            parameters=[],
            actions=[],
        )

    def test_resource(
        self,
        mock_action_role,
        mock_parameter_desc,
        mock_parameter_advanced,
        mock_stonith_parameters,
        mock_stonith_action,
        mock_stonith_port,
    ):
        mock_action_role.return_value = "from action role"
        mock_parameter_desc.return_value = "from parameter desc"
        mock_parameter_advanced.return_value = "from parameter advanced"
        mock_stonith_parameters.return_value = "from stonith parameters"
        mock_stonith_action.return_value = "from stonith action"
        mock_stonith_port.return_value = "from stonith port"

        metadata = self._fixture_metadata(
            ra.ResourceAgentName("ocf", "pacemaker", "Dummy")
        )
        self.assertEqual(
            ra.pcs_transform.ocf_unified_to_pcs(metadata),
            "from action role",
        )

        mock_action_role.assert_called_once_with(metadata)
        mock_parameter_desc.assert_not_called()
        mock_parameter_advanced.assert_not_called()
        mock_stonith_parameters.assert_not_called()
        mock_stonith_action.assert_not_called()
        mock_stonith_port.assert_not_called()

    def test_stonith(
        self,
        mock_action_role,
        mock_parameter_desc,
        mock_parameter_advanced,
        mock_stonith_parameters,
        mock_stonith_action,
        mock_stonith_port,
    ):
        mock_action_role.return_value = "from action role"
        mock_parameter_desc.return_value = "from parameter desc"
        mock_parameter_advanced.return_value = "from parameter advanced"
        mock_stonith_parameters.return_value = "from stonith parameters"
        mock_stonith_action.return_value = "from stonith action"
        mock_stonith_port.return_value = "from stonith port"

        metadata = self._fixture_metadata(
            ra.ResourceAgentName("stonith", None, "fence_xvm")
        )
        self.assertEqual(
            ra.pcs_transform.ocf_unified_to_pcs(metadata),
            "from stonith port",
        )

        mock_action_role.assert_called_once_with(metadata)
        mock_parameter_desc.assert_not_called()
        mock_parameter_advanced.assert_not_called()
        mock_stonith_parameters.assert_called_once_with("from action role")
        mock_stonith_action.assert_called_once_with("from stonith parameters")
        mock_stonith_port.assert_called_once_with("from stonith action")

    def test_pcmk_fake(
        self,
        mock_action_role,
        mock_parameter_desc,
        mock_parameter_advanced,
        mock_stonith_parameters,
        mock_stonith_action,
        mock_stonith_port,
    ):
        mock_action_role.return_value = "from action role"
        mock_parameter_desc.return_value = "from parameter desc"
        mock_parameter_advanced.return_value = "from parameter advanced"
        mock_stonith_parameters.return_value = "from stonith parameters"
        mock_stonith_action.return_value = "from stonith action"
        mock_stonith_port.return_value = "from stonith port"

        metadata = self._fixture_metadata(
            ra.ResourceAgentName(
                ra.const.FAKE_AGENT_STANDARD, None, ra.const.PACEMAKER_FENCED
            )
        )
        self.assertEqual(
            ra.pcs_transform.ocf_unified_to_pcs(metadata),
            "from parameter advanced",
        )

        mock_action_role.assert_called_once_with(metadata)
        mock_parameter_desc.assert_called_once_with("from action role")
        mock_parameter_advanced.assert_called_once_with("from parameter desc")
        mock_stonith_parameters.assert_not_called()
        mock_stonith_action.assert_not_called()
        mock_stonith_port.assert_not_called()


class MetadataActionTranslateRole(TestCase):
    @staticmethod
    def _fixture_metadata(actions):
        return ra.ResourceAgentMetadata(
            ra.ResourceAgentName("standard", "provider", "type"),
            agent_exists=True,
            ocf_version=ra.const.OCF_1_0,
            shortdesc=None,
            longdesc=None,
            parameters=[],
            actions=actions,
        )

    @staticmethod
    def _fixture_action(role, interval):
        return ra.ResourceAgentAction(
            name="monitor",
            timeout=None,
            interval=interval,
            role=role,
            start_delay=None,
            depth=None,
            automatic=False,
            on_target=False,
        )

    def test_no_actions(self):
        metadata_in = self._fixture_metadata([])
        metadata_out = self._fixture_metadata([])
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_action_translate_role(metadata_in),
            metadata_out,
        )

    @mock.patch(
        "pcs.common.const.PCMK_ROLE_PROMOTED_PRIMARY",
        const.PCMK_ROLE_PROMOTED_LEGACY,
    )
    @mock.patch(
        "pcs.common.const.PCMK_ROLE_UNPROMOTED_PRIMARY",
        const.PCMK_ROLE_UNPROMOTED_LEGACY,
    )
    def test_role_old_agent_old_preferred(self):
        metadata_in = self._fixture_metadata(
            [
                self._fixture_action(const.PCMK_ROLE_PROMOTED_LEGACY, "10"),
                self._fixture_action(const.PCMK_ROLE_UNPROMOTED_LEGACY, "11"),
            ]
        )
        metadata_out = self._fixture_metadata(
            [
                self._fixture_action(const.PCMK_ROLE_PROMOTED_LEGACY, "10"),
                self._fixture_action(const.PCMK_ROLE_UNPROMOTED_LEGACY, "11"),
            ]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_action_translate_role(metadata_in),
            metadata_out,
        )

    @mock.patch(
        "pcs.common.const.PCMK_ROLE_PROMOTED_PRIMARY",
        const.PCMK_ROLE_PROMOTED_LEGACY,
    )
    @mock.patch(
        "pcs.common.const.PCMK_ROLE_UNPROMOTED_PRIMARY",
        const.PCMK_ROLE_UNPROMOTED_LEGACY,
    )
    def test_role_new_agent_old_preferred(self):
        metadata_in = self._fixture_metadata(
            [
                self._fixture_action(const.PCMK_ROLE_PROMOTED, "10"),
                self._fixture_action(const.PCMK_ROLE_UNPROMOTED, "11"),
            ]
        )
        metadata_out = self._fixture_metadata(
            [
                self._fixture_action(const.PCMK_ROLE_PROMOTED_LEGACY, "10"),
                self._fixture_action(const.PCMK_ROLE_UNPROMOTED_LEGACY, "11"),
            ]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_action_translate_role(metadata_in),
            metadata_out,
        )

    @mock.patch(
        "pcs.common.const.PCMK_ROLE_PROMOTED_PRIMARY",
        const.PCMK_ROLE_PROMOTED,
    )
    @mock.patch(
        "pcs.common.const.PCMK_ROLE_UNPROMOTED_PRIMARY",
        const.PCMK_ROLE_UNPROMOTED,
    )
    def test_role_old_agent_new_preferred(self):
        metadata_in = self._fixture_metadata(
            [
                self._fixture_action(const.PCMK_ROLE_PROMOTED_LEGACY, "10"),
                self._fixture_action(const.PCMK_ROLE_UNPROMOTED_LEGACY, "11"),
            ]
        )
        metadata_out = self._fixture_metadata(
            [
                self._fixture_action(const.PCMK_ROLE_PROMOTED, "10"),
                self._fixture_action(const.PCMK_ROLE_UNPROMOTED, "11"),
            ]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_action_translate_role(metadata_in),
            metadata_out,
        )

    @mock.patch(
        "pcs.common.const.PCMK_ROLE_PROMOTED_PRIMARY",
        const.PCMK_ROLE_PROMOTED,
    )
    @mock.patch(
        "pcs.common.const.PCMK_ROLE_UNPROMOTED_PRIMARY",
        const.PCMK_ROLE_UNPROMOTED,
    )
    def test_role_new_agent_new_preferred(self):
        metadata_in = self._fixture_metadata(
            [
                self._fixture_action(const.PCMK_ROLE_PROMOTED, "10"),
                self._fixture_action(const.PCMK_ROLE_UNPROMOTED, "11"),
            ]
        )
        metadata_out = self._fixture_metadata(
            [
                self._fixture_action(const.PCMK_ROLE_PROMOTED, "10"),
                self._fixture_action(const.PCMK_ROLE_UNPROMOTED, "11"),
            ]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_action_translate_role(metadata_in),
            metadata_out,
        )


class MetadataParameterExtractAdvancedFromDesc(TestCase):
    advanced_str = "Advanced use only"

    @staticmethod
    def _fixture_metadata(parameters):
        return ra.ResourceAgentMetadata(
            ra.ResourceAgentName("standard", "provider", "type"),
            agent_exists=True,
            ocf_version=ra.const.OCF_1_0,
            shortdesc=None,
            longdesc=None,
            parameters=parameters,
            actions=[],
        )

    @staticmethod
    def _fixture_parameter(shortdesc, longdesc, advanced):
        return ra.ResourceAgentParameter(
            name="test-parameter",
            shortdesc=shortdesc,
            longdesc=longdesc,
            type="string",
            default=None,
            enum_values=None,
            required=False,
            advanced=advanced,
            deprecated=False,
            deprecated_by=[],
            deprecated_desc=None,
            unique_group=None,
            reloadable=False,
        )

    def test_no_parameters(self):
        metadata_in = self._fixture_metadata([])
        metadata_out = self._fixture_metadata([])
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_parameter_extract_advanced_from_desc(
                metadata_in
            ),
            metadata_out,
        )

    def test_no_shortdesc(self):
        metadata_in = self._fixture_metadata(
            [self._fixture_parameter(None, None, False)]
        )
        metadata_out = self._fixture_metadata(
            [self._fixture_parameter(None, None, False)]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_parameter_extract_advanced_from_desc(
                metadata_in
            ),
            metadata_out,
        )

    def test_no_advanced_str(self):
        metadata_in = self._fixture_metadata(
            [self._fixture_parameter("some shortdesc", None, False)]
        )
        metadata_out = self._fixture_metadata(
            [self._fixture_parameter("some shortdesc", None, False)]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_parameter_extract_advanced_from_desc(
                metadata_in
            ),
            metadata_out,
        )

    def test_advanced_str_in_shortedsc(self):
        metadata_in = self._fixture_metadata(
            [
                self._fixture_parameter(
                    f"{self.advanced_str}: some shortdesc", None, False
                )
            ]
        )
        metadata_out = self._fixture_metadata(
            [
                self._fixture_parameter(
                    f"{self.advanced_str}: some shortdesc", None, True
                )
            ]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_parameter_extract_advanced_from_desc(
                metadata_in
            ),
            metadata_out,
        )

    def test_advanced_str_in_shortdesc_end(self):
        metadata_in = self._fixture_metadata(
            [
                self._fixture_parameter(
                    f"some shortdesc {self.advanced_str}", None, False
                )
            ]
        )
        metadata_out = self._fixture_metadata(
            [
                self._fixture_parameter(
                    f"some shortdesc {self.advanced_str}", None, False
                )
            ]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_parameter_extract_advanced_from_desc(
                metadata_in
            ),
            metadata_out,
        )

    def test_advanced_str_in_longdesc(self):
        metadata_in = self._fixture_metadata(
            [
                self._fixture_parameter(
                    None, f"{self.advanced_str}: some longdesc", False
                )
            ]
        )
        metadata_out = self._fixture_metadata(
            [
                self._fixture_parameter(
                    None, f"{self.advanced_str}: some longdesc", False
                )
            ]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_parameter_extract_advanced_from_desc(
                metadata_in
            ),
            metadata_out,
        )


class MetadataParameterJoinShortLongDesc(TestCase):
    @staticmethod
    def _fixture_metadata(parameters):
        return ra.ResourceAgentMetadata(
            ra.ResourceAgentName("standard", "provider", "type"),
            agent_exists=True,
            ocf_version=ra.const.OCF_1_0,
            shortdesc=None,
            longdesc=None,
            parameters=parameters,
            actions=[],
        )

    @staticmethod
    def _fixture_parameter(shortdesc, longdesc):
        return ra.ResourceAgentParameter(
            name="test-parameter",
            shortdesc=shortdesc,
            longdesc=longdesc,
            type="string",
            default=None,
            enum_values=None,
            required=False,
            advanced=False,
            deprecated=False,
            deprecated_by=[],
            deprecated_desc=None,
            unique_group=None,
            reloadable=False,
        )

    def test_no_parameters(self):
        metadata_in = self._fixture_metadata([])
        metadata_out = self._fixture_metadata([])
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_parameter_join_short_long_desc(
                metadata_in
            ),
            metadata_out,
        )

    def test_no_shortdesc_no_longdesc(self):
        metadata_in = self._fixture_metadata(
            [self._fixture_parameter(None, None)]
        )
        metadata_out = self._fixture_metadata(
            [self._fixture_parameter(None, None)]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_parameter_join_short_long_desc(
                metadata_in
            ),
            metadata_out,
        )

    def test_shortdesc_only(self):
        metadata_in = self._fixture_metadata(
            [self._fixture_parameter("shortdesc", None)]
        )
        metadata_out = self._fixture_metadata(
            [self._fixture_parameter("shortdesc", None)]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_parameter_join_short_long_desc(
                metadata_in
            ),
            metadata_out,
        )

    def test_longdesc_only(self):
        metadata_in = self._fixture_metadata(
            [self._fixture_parameter(None, "longdesc")]
        )
        metadata_out = self._fixture_metadata(
            [self._fixture_parameter(None, "longdesc")]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_parameter_join_short_long_desc(
                metadata_in
            ),
            metadata_out,
        )

    def test_shortdesc_ang_longdesc(self):
        metadata_in = self._fixture_metadata(
            [self._fixture_parameter("shortdesc", "longdesc")]
        )
        metadata_out = self._fixture_metadata(
            [self._fixture_parameter("shortdesc", "shortdesc\nlongdesc")]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_parameter_join_short_long_desc(
                metadata_in
            ),
            metadata_out,
        )


class MetadataRemoveUnwantedStonithParameters(TestCase):
    @staticmethod
    def _fixture_metadata(parameters):
        return ra.ResourceAgentMetadata(
            ra.ResourceAgentName("standard", "provider", "type"),
            agent_exists=True,
            ocf_version=ra.const.OCF_1_0,
            shortdesc=None,
            longdesc=None,
            parameters=parameters,
            actions=[],
        )

    @staticmethod
    def _fixture_parameter(name):
        return ra.ResourceAgentParameter(
            name=name,
            shortdesc=None,
            longdesc=None,
            type="string",
            default=None,
            enum_values=None,
            required=False,
            advanced=False,
            deprecated=False,
            deprecated_by=[],
            deprecated_desc=None,
            unique_group=None,
            reloadable=False,
        )

    def test_no_parameters(self):
        metadata_in = self._fixture_metadata([])
        metadata_out = self._fixture_metadata([])
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_remove_unwanted_stonith_parameters(
                metadata_in
            ),
            metadata_out,
        )

    def test_success(self):
        metadata_in = self._fixture_metadata(
            [
                self._fixture_parameter("param1"),
                self._fixture_parameter("help"),
                self._fixture_parameter("param2"),
                self._fixture_parameter("version"),
            ]
        )
        metadata_out = self._fixture_metadata(
            [
                self._fixture_parameter("param1"),
                self._fixture_parameter("param2"),
            ]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_remove_unwanted_stonith_parameters(
                metadata_in
            ),
            metadata_out,
        )


class MetadataMakeStonithActionParameterDeprecated(TestCase):
    @staticmethod
    def _fixture_metadata(parameters):
        return ra.ResourceAgentMetadata(
            ra.ResourceAgentName("standard", "provider", "type"),
            agent_exists=True,
            ocf_version=ra.const.OCF_1_0,
            shortdesc=None,
            longdesc=None,
            parameters=parameters,
            actions=[],
        )

    @staticmethod
    def _fixture_parameter(name, required, advanced, deprecated, deprecated_by):
        return ra.ResourceAgentParameter(
            name,
            shortdesc=None,
            longdesc=None,
            type="string",
            default=None,
            enum_values=None,
            required=required,
            advanced=advanced,
            deprecated=deprecated,
            deprecated_by=deprecated_by,
            deprecated_desc=None,
            unique_group=None,
            reloadable=False,
        )

    def test_no_parameters(self):
        metadata_in = self._fixture_metadata([])
        metadata_out = self._fixture_metadata([])
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_make_stonith_action_parameter_deprecated(
                metadata_in
            ),
            metadata_out,
        )

    def test_no_action_parameter(self):
        metadata_in = self._fixture_metadata(
            [
                self._fixture_parameter(
                    "monitor",
                    required=True,
                    advanced=False,
                    deprecated=False,
                    deprecated_by=[],
                )
            ]
        )
        metadata_out = self._fixture_metadata(
            [
                self._fixture_parameter(
                    "monitor",
                    required=True,
                    advanced=False,
                    deprecated=False,
                    deprecated_by=[],
                )
            ]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_make_stonith_action_parameter_deprecated(
                metadata_in
            ),
            metadata_out,
        )

    def test_action_parameter(self):
        metadata_in = self._fixture_metadata(
            [
                self._fixture_parameter(
                    "action",
                    required=True,
                    advanced=False,
                    deprecated=False,
                    deprecated_by=["new-action"],
                )
            ]
        )
        metadata_out = self._fixture_metadata(
            [
                self._fixture_parameter(
                    "action",
                    required=False,
                    advanced=True,
                    deprecated=True,
                    deprecated_by=(
                        ra.const.STONITH_ACTION_REPLACED_BY + ["new-action"]
                    ),
                )
            ]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_make_stonith_action_parameter_deprecated(
                metadata_in
            ),
            metadata_out,
        )


class MetadataMakeStonithPortParameterNotRequired(TestCase):
    @staticmethod
    def _fixture_metadata(parameters):
        return ra.ResourceAgentMetadata(
            ra.ResourceAgentName("standard", "provider", "type"),
            agent_exists=True,
            ocf_version=ra.const.OCF_1_0,
            shortdesc=None,
            longdesc=None,
            parameters=parameters,
            actions=[],
        )

    @staticmethod
    def _fixture_parameter(name, required, deprecated_by):
        return ra.ResourceAgentParameter(
            name,
            shortdesc=None,
            longdesc=None,
            type="string",
            default=None,
            enum_values=None,
            required=required,
            advanced=False,
            deprecated=bool(deprecated_by),
            deprecated_by=deprecated_by,
            deprecated_desc=None,
            unique_group=None,
            reloadable=False,
        )

    def test_no_parameters(self):
        metadata_in = self._fixture_metadata([])
        metadata_out = self._fixture_metadata([])
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_make_stonith_port_parameter_not_required(
                metadata_in
            ),
            metadata_out,
        )

    def test_no_port_parameter(self):
        metadata_in = self._fixture_metadata(
            [self._fixture_parameter("param", True, [])]
        )
        metadata_out = self._fixture_metadata(
            [self._fixture_parameter("param", True, [])]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_make_stonith_port_parameter_not_required(
                metadata_in
            ),
            metadata_out,
        )

    def test_modify_port_parameter(self):
        metadata_in = self._fixture_metadata(
            [self._fixture_parameter("port", True, [])]
        )
        metadata_out = self._fixture_metadata(
            [self._fixture_parameter("port", False, [])]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_make_stonith_port_parameter_not_required(
                metadata_in
            ),
            metadata_out,
        )

    def test_modify_port_and_deprecations(self):
        metadata_in = self._fixture_metadata(
            [
                self._fixture_parameter("old-port", True, ["port"]),
                self._fixture_parameter("port", True, ["new-port"]),
                self._fixture_parameter(
                    "new-port", True, ["new-port2a", "new-port2b"]
                ),
                self._fixture_parameter("new-port2a", True, []),
                self._fixture_parameter("new-port2b", True, []),
            ]
        )
        metadata_out = self._fixture_metadata(
            [
                self._fixture_parameter("old-port", True, ["port"]),
                self._fixture_parameter("port", False, ["new-port"]),
                self._fixture_parameter(
                    "new-port", False, ["new-port2a", "new-port2b"]
                ),
                self._fixture_parameter("new-port2a", False, []),
                self._fixture_parameter("new-port2b", False, []),
            ]
        )
        self.assertEqual(
            # pylint: disable=protected-access
            ra.pcs_transform._metadata_make_stonith_port_parameter_not_required(
                metadata_in
            ),
            metadata_out,
        )



from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from dataclasses import dataclass

from pcs.common.interface.dto import (
    ImplementsFromDto,
    ImplementsToDto,
)

from .dto import (
    ReportItemContextDto,
    ReportItemDto,
    ReportItemMessageDto,
    ReportItemSeverityDto,
)
from .types import (
    ForceCode,
    MessageCode,
    SeverityLevel,
)


@dataclass(frozen=True)
class ReportItemSeverity(ImplementsToDto, ImplementsFromDto):
    # pylint: disable=invalid-name
    ERROR = SeverityLevel("ERROR")
    WARNING = SeverityLevel("WARNING")
    INFO = SeverityLevel("INFO")
    DEBUG = SeverityLevel("DEBUG")

    level: SeverityLevel
    force_code: Optional[ForceCode] = None

    def to_dto(self) -> ReportItemSeverityDto:
        return ReportItemSeverityDto(
            level=self.level,
            force_code=self.force_code,
        )

    @classmethod
    def from_dto(cls, dto_obj: ReportItemSeverityDto) -> "ReportItemSeverity":
        return cls(
            level=dto_obj.level,
            force_code=dto_obj.force_code,
        )

    @classmethod
    def error(
        cls, force_code: Optional[ForceCode] = None
    ) -> "ReportItemSeverity":
        return cls(level=cls.ERROR, force_code=force_code)

    @classmethod
    def warning(cls) -> "ReportItemSeverity":
        return cls(level=cls.WARNING)

    @classmethod
    def info(cls) -> "ReportItemSeverity":
        return cls(level=cls.INFO)

    @classmethod
    def debug(cls) -> "ReportItemSeverity":
        return cls(level=cls.DEBUG)


def get_severity(
    force_code: Optional[ForceCode], is_forced: bool
) -> ReportItemSeverity:
    if is_forced:
        return ReportItemSeverity(ReportItemSeverity.WARNING)
    return ReportItemSeverity(ReportItemSeverity.ERROR, force_code)


@dataclass(frozen=True, init=False)
class ReportItemMessage(ImplementsToDto):
    _code = MessageCode("")

    @property
    def message(self) -> str:
        raise NotImplementedError()

    @property
    def code(self) -> MessageCode:
        return self._code

    def to_dto(self) -> ReportItemMessageDto:
        payload: Dict[str, Any] = {}
        if hasattr(self.__class__, "__annotations__"):
            try:
                annotations = self.__class__.__annotations__
            except AttributeError as e:
                raise AssertionError() from e
            for attr_name in annotations.keys():
                if attr_name.startswith("_") or attr_name in ("message",):
                    continue
                attr_val = getattr(self, attr_name)
                if hasattr(attr_val, "to_dto"):
                    payload[attr_name] = attr_val.to_dto()
                else:
                    payload[attr_name] = attr_val

        return ReportItemMessageDto(
            code=self.code,
            message=self.message,
            payload=payload,
        )


@dataclass(frozen=True)
class ReportItemContext(ImplementsToDto, ImplementsFromDto):
    node: str

    @classmethod
    def from_dto(cls, dto_obj: ReportItemContextDto) -> "ReportItemContext":
        return cls(node=dto_obj.node)

    def to_dto(self) -> ReportItemContextDto:
        return ReportItemContextDto(node=self.node)


@dataclass
class ReportItem(ImplementsToDto):
    severity: ReportItemSeverity
    message: ReportItemMessage
    context: Optional[ReportItemContext] = None

    @classmethod
    def error(
        cls,
        message: ReportItemMessage,
        force_code: Optional[ForceCode] = None,
        context: Optional[ReportItemContext] = None,
    ) -> "ReportItem":
        return cls(
            severity=ReportItemSeverity.error(force_code),
            message=message,
            context=context,
        )

    @classmethod
    def warning(
        cls,
        message: ReportItemMessage,
        context: Optional[ReportItemContext] = None,
    ) -> "ReportItem":
        return cls(
            severity=ReportItemSeverity.warning(),
            message=message,
            context=context,
        )

    @classmethod
    def info(
        cls,
        message: ReportItemMessage,
        context: Optional[ReportItemContext] = None,
    ) -> "ReportItem":
        return cls(
            severity=ReportItemSeverity.info(),
            message=message,
            context=context,
        )

    @classmethod
    def debug(
        cls,
        message: ReportItemMessage,
        context: Optional[ReportItemContext] = None,
    ) -> "ReportItem":
        return cls(
            severity=ReportItemSeverity.debug(),
            message=message,
            context=context,
        )

    def to_dto(self) -> ReportItemDto:
        return ReportItemDto(
            severity=self.severity.to_dto(),
            context=self.context.to_dto() if self.context else None,
            message=self.message.to_dto(),
        )


ReportItemList = List[ReportItem]
